import io
from importlib import resources as impresources
from typing import Literal

from PIL import Image
from pypdf import PageObject, PdfReader

QrSupported = Literal["erkc", "kapremont", "peni"]

# img = impresources.files("erkc63") / "paid.png"

# _PAID_LOGO = Image.open(img.name).convert("RGBA")


def _paid_logo(size: float) -> Image.Image:
    img = _PAID_LOGO.copy()
    img.thumbnail((size, size), Image.Resampling.BICUBIC)

    return img


def _img_to_png(img: Image.Image) -> bytes:
    bio = io.BytesIO()
    img = img.convert("P", palette=Image.Palette.WEB)
    img.save(bio, format="png", optimize=True)

    return bio.getvalue()


def _img_paid(img_data: bytes, paid_scale: float) -> bytes:
    img = Image.open(io.BytesIO(img_data)).convert("RGB")
    # Resize the logo to logo_max_size
    logo = _paid_logo(min(img.width, img.height) * paid_scale)
    # Calculate the center of the QR code
    box = (img.width - logo.width) // 2, (img.height - logo.height) // 2
    img.paste(logo, box, logo)

    return _img_to_png(img)


def _page_img(page: PageObject, name: str) -> bytes:
    for img in page.images:
        if img.name == name:
            return img.data

    raise FileNotFoundError("Image %s not found.", name)


class QrCodes:
    _codes: dict[QrSupported, bytes]
    _paid_scale: float

    def __init__(
        self, pdf_erkc: bytes, pdf_peni: bytes, paid_scale: float = 0.65
    ) -> None:
        assert 0 < paid_scale <= 1

        self._paid_scale = paid_scale
        self._codes = {}

        if pdf_erkc:
            page = PdfReader(io.BytesIO(pdf_erkc)).pages[0]
            self._codes["erkc"] = _page_img(page, "img2.png")
            self._codes["kapremont"] = _page_img(page, "img4.png")

        if pdf_peni:
            page = PdfReader(io.BytesIO(pdf_peni)).pages[0]
            self._codes["peni"] = _page_img(page, "img0.png")

    def qr(self, qr: QrSupported, paid: bool = False) -> bytes | None:
        if img := self._codes.get(qr):
            return _img_paid(img, self._paid_scale) if paid else img

    def erkc(self, is_paid: bool = False) -> bytes | None:
        """QR-код оплаты коммунальных услуг."""

        return self.qr("erkc", is_paid)

    def kapremont(self, is_paid: bool = False) -> bytes | None:
        """QR-код оплаты капитального ремонта."""

        return self.qr("kapremont", is_paid)

    def peni(self, is_paid: bool = False) -> bytes | None:
        """QR-код оплаты пени."""

        return self.qr("peni", is_paid)



import datetime as dt
import re
from typing import Any


def date_first_day(_date: dt.date) -> dt.date:
    """Возвращает дату первого числа месяца."""

    return dt.date(_date.year, _date.month, 1)


def date_last_accrual(accrual_day: int = 25) -> dt.date:
    """Возвращает дату последнего расчетного периода."""

    if (today := dt.date.today()).day > accrual_day:
        return dt.date(today.year, today.month, 1)

    if today.month != 1:
        return dt.date(today.year, today.month - 1, 1)

    return dt.date(today.year - 1, 12, 1)


def first_int(_str: str) -> int:
    """Возвращает первое целое число в строке."""

    for idx, sym in enumerate(_str):
        if not sym.isdigit():
            _str = _str[:idx]
            break

    return int(_str)


def to_float(_str: Any) -> float:
    """Преобразует строку в число."""

    return float(str(_str).replace(",", ".").replace(" ", ""))


def str_to_date(_str: str) -> dt.date:
    """Преобразует строку вида `dd.mm.yy` в дату."""

    return dt.datetime.strptime(_str, "%d.%m.%y").date()


def data_attr(_str: str) -> str:
    """Извлекает строку из атрибута данных тэга."""

    if m := re.search(r' data-\w+="([\w/.+=]+)"', _str):
        return m.group(1)

    raise ValueError


def date_attr(_str: str) -> dt.date:
    return str_to_date(data_attr(_str))


def date_to_str(_date: dt.date) -> str:
    """Преобразует дату в строку вида `dd.mm.YYYY`."""

    return _date.strftime("%d.%m.%Y")


def str_normalize(_str: str) -> str:
    """Нормализует строку, удаляя лишние пробелы."""

    return " ".join(_str.split())



import dataclasses as dc


@dc.dataclass
class PublicAccountInfo:
    """Открытая информация о лицевом счете."""

    account: int
    """Номер лицевого счета"""
    address: str
    """Адрес"""
    balance: float
    """Задолженность"""
    peni: float
    """Пени"""

    def __repr__(self) -> str:
        return (
            f"Лицевой счет:  {self.account}\n"
            f"Адрес:         {self.address}\n"
            f"Задолженность: {self.balance}\n"
            f"Пени:          {self.peni}\n"
        )


@dc.dataclass
class AccountInfo:
    """Информация о лицевом счете"""

    address: str
    """Адрес"""
    person: str
    """Собственник"""
    phone: str
    """Телефон"""
    email: str
    """Электронная почта"""
    account: int
    """Лицевой счет"""
    square: float
    """Общая площадь"""
    registered: int
    """Зарегистрировано"""
    hosted: int
    """Проживает"""
    document: str
    """Право собственности"""



import asyncio
import json
import logging

from .client import ErkcClient

logging.basicConfig(level=logging.DEBUG)

with open("secrets.json") as f:
    secrets: dict[str, str] = json.load(f)


async def main():
    async with ErkcClient(secrets["login"], secrets["password"]) as cli:
        print(await cli.account_info())
        print(await cli.meters_info())

        for m in await cli.meters_history():
            for value in m.history:
                print(value)

        for x in await cli.year_accruals(include_details=True):
            print(x)

        for x in await cli.accruals_history():
            print(x)

        for x in await cli.payments_history():
            print(x)


asyncio.run(main())



from .client import ErkcClient
from .errors import (
    AccountBindingError,
    AccountNotFound,
    ApiError,
    AuthorizationError,
    AuthorizationRequired,
    ErkcError,
    ParsingError,
)

__all__ = [
    "ErkcClient",
    "ErkcError",
    "ApiError",
    "ParsingError",
    "AuthorizationError",
    "AccountBindingError",
    "AuthorizationRequired",
    "AccountNotFound",
]



class ErkcError(Exception):
    pass


class ParsingError(ErkcError):
    pass


class ApiError(ErkcError):
    pass


class AccountBindingError(ApiError):
    pass


class AuthorizationError(ApiError):
    pass


class AuthorizationRequired(ApiError):
    pass


class AccountNotFound(ApiError):
    pass



import dataclasses as dc
import datetime as dt


@dc.dataclass(frozen=True)
class Payment:
    """
    Платеж.

    Объект ответа на запрос `paymentsHistory`.
    """

    date: dt.date
    """Дата"""
    summa: float
    """Сумма"""
    provider: str
    """Платежный провайдер"""



import dataclasses as dc
import datetime as dt
from typing import Mapping

from .errors import ErkcError


@dc.dataclass(frozen=True)
class AccrualDetalization:
    """Детализация услуги"""

    tariff: float
    """Тариф"""
    saldo_in: float
    """Входящее сальдо (долг на начало расчетного периода)"""
    billed: float
    """Начислено"""
    reee: float
    """Перерасчет"""
    quality: float
    """Снято за качество"""
    payment: float
    """Платеж"""
    saldo_out: float
    """Исходящее сальдо (долг на конец расчетного периода)"""
    volume: float
    """Объем"""


@dc.dataclass
class Accrual:
    """
    Квитанция.

    Объект ответа на запрос `getReceipts`.
    """

    account: int
    """Лицевой счет"""
    date: dt.date
    """Дата формирования"""
    summa: float
    """Сумма"""
    peni: float
    """Пени"""
    bill_id: str | None = None
    """Идентификатор квитанции для скачивания"""
    peni_id: str | None = None
    """Идентификатор квитанции на пени для скачивания"""
    details: Mapping[str, AccrualDetalization] | None = None
    """Детализация услуг"""

    def _sum(self, attr: str) -> float:
        if self.details:
            return sum(getattr(x, attr) for x in self.details.values())

        raise ErkcError("Отсутствует детализация по услугам")

    @property
    def saldo_in(self) -> float:
        """Входящее сальдо (долг на начало расчетного периода)"""
        return self._sum("saldo_in")

    @property
    def billed(self) -> float:
        """Начислено"""
        return self._sum("billed")

    @property
    def reee(self) -> float:
        """Перерасчет"""
        return self._sum("reee")

    @property
    def quality(self) -> float:
        """Снято за качество"""
        return self._sum("quality")

    @property
    def payment(self) -> float:
        """Платеж"""
        return self._sum("payment")

    @property
    def saldo_out(self) -> float:
        """Исходящее сальдо (долг на конец расчетного периода)"""
        return self._sum("saldo_out")

    @property
    def is_correct(self) -> bool:
        """Корректен (сумма счета совпадает с суммой начислений по услугам)"""
        return self.summa == self.billed

    @property
    def is_paid(self) -> bool:
        """Оплачен"""
        return not self.saldo_out

    @property
    def tariffs(self):
        assert self.details
        return {k: v.tariff for k, v in self.details.items()}


@dc.dataclass
class MonthAccrual:
    """
    Начисление.

    Объект ответа на запрос `accrualsHistory`.
    """

    account: int
    """Лицевой счет"""
    date: dt.date
    """Дата"""
    saldo_in: float
    """Входящее сальдо (долг на начало расчетного периода)"""
    summa: float
    """Начислено"""
    payment: float
    """Платеж"""
    saldo_out: float
    """Исходящее сальдо (долг на конец расчетного периода)"""
    details: Mapping[str, AccrualDetalization] | None = None
    """Детализация услуг"""


Accruals = Accrual | MonthAccrual



import dataclasses as dc
import datetime as dt


@dc.dataclass(frozen=True)
class MeterInfo:
    name: str
    """Ресурс учета"""
    serial: str
    """Серийный номер"""

    def __eq__(self, other: "MeterInfo") -> bool:
        return self.name == other.name and self.serial == other.serial


@dc.dataclass(frozen=True, eq=False)
class PublicMeterInfo(MeterInfo):
    """
    Информация о приборе учета.

    Результат парсинга HTML-страницы.
    """

    date: dt.date
    """Дата последнего показания"""
    value: float
    """Последнее показание"""


@dc.dataclass(frozen=True)
class MeterValue:
    """Показание счетчика"""

    date: dt.date
    """Дата"""
    value: float
    """Значение"""
    consumption: float
    """Расход"""
    source: str
    """Источник"""


@dc.dataclass(frozen=True, eq=False)
class MeterInfoHistory(MeterInfo):
    """Счетчик с архивом показаний"""

    history: tuple[MeterValue, ...]
    """Архив показаний"""



from __future__ import annotations

import asyncio
import datetime as dt
import functools
import logging
from typing import (
    Any,
    Awaitable,
    Callable,
    Concatenate,
    Coroutine,
    Iterable,
    Mapping,
    ParamSpec,
    Self,
    Sequence,
    TypeVar,
)

import aiohttp
import yarl

from .account import AccountInfo, PublicAccountInfo
from .accrual import Accrual, AccrualDetalization, Accruals, MonthAccrual
from .bills import QrCodes
from .errors import (
    AccountBindingError,
    AccountNotFound,
    AuthorizationError,
    AuthorizationRequired,
    ParsingError,
)
from .meters import MeterInfoHistory, MeterValue, PublicMeterInfo
from .parsers import parse_account, parse_accounts, parse_meters, parse_token
from .payment import Payment
from .utils import (
    data_attr,
    date_attr,
    date_last_accrual,
    date_to_str,
    str_normalize,
    str_to_date,
    to_float,
)

_LOGGER = logging.getLogger(__name__)

_MIN_DATE = dt.date(2018, 1, 1)
_MAX_DATE = dt.date(2099, 12, 31)

_BASE_URL = yarl.URL("https://lk.erkc63.ru")

P = ParamSpec("P")
T = TypeVar("T")

ClientMethod = Callable[Concatenate["Self@ErkcClient", P], Awaitable[T]]


def api(
    *,
    auth_required: bool = True,
    public: bool = False,
) -> Callable[[ClientMethod], ClientMethod]:
    """Декоратор методов API клиента"""

    def decorator(func: ClientMethod):
        @functools.wraps(func)
        async def _wrapper(self: "ErkcClient", *args, **kwargs):
            if public and self.authorized:
                raise AuthorizationRequired("Требуется выход из аккаунта")

            if auth_required and not self.authorized:
                raise AuthorizationRequired("Требуется авторизация")

            return await func(self, *args, **kwargs)

        return _wrapper

    return decorator


class ErkcClient:
    """
    Клиент личного кабинета ЕРКЦ.
    """

    _cli: aiohttp.ClientSession
    """Клиентская сессия."""
    _login: str | None
    """Логин (адрес электронной почты)."""
    _password: str | None
    """Пароль."""
    _token: str | None
    """Токен сессии."""
    _accounts: tuple[int, ...] | None
    """Лицевые счета, привязанные к аккаунту."""

    def __init__(
        self,
        login: str | None = None,
        password: str | None = None,
        *,
        session: aiohttp.ClientSession | None = None,
        close_connector: bool = True,
    ) -> None:
        """
        Создает клиент личного кабинета ЕРКЦ.

        Параметры:
        - `login`: логин (электронная почта).
        - `password`: пароль.
        - `session`: готовый объект `aiohttp.ClientSession`.
        - `close_connector`: закрывать коннектор при закрытии сессии.
        если он не указан в параметре `session`.
        """

        self._cli = session or aiohttp.ClientSession()
        self._login = login
        self._password = password
        self._accounts = None
        self._token = None
        self._close_connector = close_connector

    async def __aenter__(self):
        try:
            await self.open()

        except Exception:
            await self.close()
            raise

        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.close()

    def _post(self, path: str, **data: Any):
        data["_token"] = self._token
        return self._cli.post(_BASE_URL.joinpath(path), data=data)

    def _get(self, path: str, **params: Any):
        return self._cli.get(_BASE_URL.joinpath(path), params=params)

    async def _ajax(self, func: str, account: int | None, **params: Any) -> Any:
        async with self._get(f"ajax/{self._account(account)}/{func}", **params) as x:
            return await x.json()

    def _history(
        self, what: str, account: int | None, start: dt.date, end: dt.date
    ) -> Coroutine[Any, Any, Sequence[Sequence[str]]]:
        params = {"from": date_to_str(start), "to": date_to_str(end)}
        return self._ajax(f"{what}History", account, **params)

    def _update_token(self, html: str) -> None:
        self._token = parse_token(html)

    def _update_accounts(self, html: str) -> None:
        self._accounts = parse_accounts(html)

    @property
    def closed(self) -> bool:
        """Коннектор клиента закрыт."""

        return self._cli.closed

    @property
    def opened(self) -> bool:
        """Сессия открыта."""

        return self._token is not None

    @property
    def authorized(self) -> bool:
        """Авторизация в аккаунте выполнена."""

        return self._accounts is not None

    @property
    def accounts(self) -> tuple[int, ...]:
        """Лицевые счета, привязанные к аккаунту личного кабиента."""

        if self._accounts is None:
            raise AuthorizationRequired("Требуется авторизация")

        return self._accounts

    @property
    def account(self) -> int:
        """Основной лицевой счет личного кабинета."""

        if x := self.accounts:
            return x[0]

        raise AccountNotFound("Основной лицевой счет не привязан")

    def _account(self, account: int | None) -> int:
        if account is None:
            return self.account

        if account in self.accounts:
            return account

        raise AccountNotFound("Лицевой счет %d не привязан", account)

    async def open(
        self,
        login: str | None = None,
        password: str | None = None,
        auth: bool = True,
    ) -> None:
        """Открытие сессии"""

        if not self.opened:
            async with self._get("login") as x:
                self._update_token(await x.text())

        if not auth or self.authorized:
            return

        login, password = login or self._login, password or self._password

        if not (login and password):
            raise AuthorizationError("Не заданы параметры входа")

        _LOGGER.debug("Вход в аккаунт %s", login)

        async with self._post("login", login=login, password=password) as x:
            if x.url == x.history[0].url:
                raise AuthorizationError("Ошибка входа. Проверьте логин и пароль")

            self._update_accounts(await x.text())

        _LOGGER.debug("Вход в аккаунт %s успешно выполнен", login)

        # Сохраняем актуальную пару логин-пароль
        self._login, self._password = login, password

    async def close(self, close_connector: bool | None = None) -> None:
        """Выход из аккаунта личного кабинета и закрытие сессии."""

        try:
            if self.authorized:
                _LOGGER.debug("Выход из аккаунта %s", self._login)

                async with self._get("logout") as x:
                    # выход из аккаунта выполняет редирект на
                    # страницу входа с новым токеном сессии
                    self._update_token(await x.text())

                self._accounts = None

        finally:
            if close_connector or self._close_connector:
                await self._cli.close()
                self._token = None

    @api(auth_required=True)
    async def download_pdf(self, accrual: Accrual, peni: bool = False) -> bytes:
        """
        Загружает квитанцию в формате PDF. При неудаче возвращает пустые данные.

        Параметры:
        - `accrual`: квитанция.
        - `peni`: нужна квитанция пени.
        """

        if not (id := accrual.peni_id if peni else accrual.bill_id):
            return b""

        try:
            json = await self._ajax("getReceipt", accrual.account, receiptId=id)

        except Exception:
            return b""

        async with self._get(json["file"]) as x:
            return await x.read()

    @api(auth_required=True)
    async def qr_codes(self, accrual: Accrual) -> QrCodes:
        """
        Загружает PDF квитанции и извлекает QR коды оплаты.
        Возвращает объект `QrCodes`.

        Параметры:
        - `accrual`: квитанция.
        """

        result = await asyncio.gather(
            self.download_pdf(accrual, False),
            self.download_pdf(accrual, True),
        )

        return QrCodes(*result)

    @api(auth_required=True)
    async def year_accruals(
        self,
        year: int | None = None,
        *,
        account: int | None = None,
        limit: int | None = None,
        include_details: bool = False,
    ) -> tuple[Accrual, ...]:
        """
        Запрос квитанций лицевого счета за год.

        Если год не уточняется - используется текущий.

        Параметры:
        - `year`: год.
        - `account`: номер лицевого счета. Если `None` - будет использоваться
        основной лицевой счет личного кабинета.
        - `limit`: кол-во последних квитанций в ответе. По-умолчанию все квитанции за год.
        - `include_details`: дополнительный запрос детализированных затрат на каждую
        квитанцию в полученном результате. По-умолчанию: `False`.
        """

        account = self._account(account)

        resp: Sequence[Sequence[str]] = await self._ajax(
            "getReceipts", account, year=year or date_last_accrual().year
        )

        db: dict[dt.date, Accrual] = {}

        for data in resp:
            date = date_attr(data[0])

            if limit and limit == len(db) and date not in db:
                break

            record = db.setdefault(
                date,
                Accrual(
                    account=account,
                    date=date,
                    summa=to_float(data[1]),
                    peni=to_float(data[2]),
                ),
            )

            id = data_attr(data[5])

            match data[3]:
                case "общая":
                    record.bill_id = id
                case "пени":
                    record.peni_id = id
                case _:
                    raise ParsingError

        result = tuple(db.values())

        if include_details:
            await self.update_accruals(result)

        return result

    @api(auth_required=True)
    async def update_accrual(self, accrual: Accruals) -> None:
        """
        Обновление детализированных данных квитанции или начисления.

        Параметры:
        - `accrual`: квитанция/начисление для обновления.
        """

        resp: list[list[str]] = await self._ajax(
            "accrualsDetalization",
            accrual.account,
            month=accrual.date.strftime("01.%m.%y"),
        )

        accrual.details = {
            str_normalize(x[0]): AccrualDetalization(*map(to_float, x[1:]))
            for x in resp
        }

    @api(auth_required=True)
    def update_accruals(self, accruals: Iterable[Accruals]):
        """
        Обновление детализированных данных квитанций или начислений.

        Параметры:
        - `accruals`: квитанции/начисления для обновления.
        """

        return asyncio.gather(*map(self.update_accrual, accruals))

    @api(auth_required=True)
    async def meters_history(
        self,
        *,
        start: dt.date | None = None,
        end: dt.date | None = None,
        account: int | None = None,
    ) -> tuple[MeterInfoHistory, ...]:
        """
        Запрос счетчиков лицевого счета с историей показаний.

        Если даты не уточняются - результат будет включать все доступные показания.

        Параметры:
        - `start`: дата начала периода.
        - `end`: дата окончания периода (включается в ответ).
        - `account`: номер лицевого счета. Если `None` - будет использоваться
        основной лицевой счет личного кабинета.
        """

        start, end = start or _MIN_DATE, end or _MAX_DATE

        assert start <= end

        db: dict[tuple[str, str], list[MeterValue]] = {}

        while True:
            history = await self._history("counters", account, start, end)

            # Лимит записей ответа сервера - 25. Контроль превышения на случай изменения API.
            assert (num := len(history)) <= 25

            # Множество для проверки содержания в ответе данных от одной даты.
            unique_dates = set()

            for _, key, date, value, consumption, source in history:
                unique_dates.add(end := str_to_date(date[27:35]))

                # игнорируем записи без потребления
                if not (consumption := float(consumption)):
                    continue

                value = MeterValue(end, float(value), consumption, source)

                name, serial = key.split(", счетчик №", 1)
                db.setdefault((name, serial), []).append(value)

            if num < 25:
                break

            # Возможен баг: если в один день число записей больше лимита,
            # то сервер не сможет вернуть полный результат ни при каких условиях.
            # Этот случай крайне маловероятен, но выполнена проверка и обход ситуации.
            if len(unique_dates) == 1:
                _LOGGER.warning("Результат может содержать неполные данные.")

                if start == end:
                    break

                end -= dt.timedelta(days=1)
                _LOGGER.warning("Применен обход.")

        # Исключаем дублирование записей из наложенных ответов и конвертируем в кортеж
        return tuple(
            MeterInfoHistory(*k, tuple(dict.fromkeys(v))) for k, v in db.items()
        )

    @api(auth_required=True)
    async def accruals_history(
        self,
        *,
        start: dt.date | None = None,
        end: dt.date | None = None,
        account: int | None = None,
        include_details: bool = False,
    ) -> tuple[MonthAccrual, ...]:
        """
        Запрос начислений за заданный период.

        Если даты не уточняются - результат будет включать все доступные показания.

        Параметры:
        - `start`: дата начала периода.
        - `end`: дата окончания периода (включается в ответ).
        - `account`: номер лицевого счета. Если `None` - будет использоваться
        основной лицевой счет личного кабинета.
        - `include_details`: дополнительный запрос детализированных затрат на каждое
        начисление в полученном результате. По-умолчанию: `False`.
        """

        account = self._account(account)
        start, end = start or _MIN_DATE, end or _MAX_DATE

        assert start <= end

        resp = await self._history("accruals", account, start, end)

        result = []

        for date, *floats in resp:
            floats: Any = map(to_float, floats)
            accrual = MonthAccrual(account, date_attr(date), *floats)

            # запрос поломан. возвращает нулевые начисления в невалидном диапазоне дат.
            # при первом нулевом начислении прерываем цикл, так как далее все начисления тоже нулевые.
            if not accrual.summa:
                break

            result.append(accrual)

        if include_details:
            await self.update_accruals(result)

        return tuple(result)

    @api(auth_required=True)
    async def payments_history(
        self,
        *,
        start: dt.date | None = None,
        end: dt.date | None = None,
        account: int | None = None,
    ) -> tuple[Payment, ...]:
        """
        Запрос истории платежей за заданный период.

        Если даты не уточняются - результат будет включать все доступные показания.

        Параметры:
        - `start`: дата начала периода.
        - `end`: дата окончания периода (включается в ответ).
        - `account`: номер лицевого счета. Если `None` - будет использоваться
        основной лицевой счет личного кабинета.
        """

        start, end = start or _MIN_DATE, end or _MAX_DATE

        assert start <= end

        resp = await self._history("payments", account, start, end)
        result = (Payment(date_attr(x), to_float(y), z) for x, y, z in resp)

        # Ответ содержит нулевые платежи (внутренние перерасчеты). Применим фильтр.
        return tuple(x for x in result if x.summa)

    @api(auth_required=True)
    async def account_info(self, account: int | None = None) -> AccountInfo:
        """
        Запрос информации о лицевом счете.

        Параметры:
        - `account`: номер лицевого счета. Если `None` - будет использоваться
        основной лицевой счет личного кабинета.
        """

        account = self._account(account)

        async with self._get(f"account/{account}") as x:
            return parse_account(await x.text())

    @api(auth_required=True)
    async def account_add(
        self,
        account: int | PublicAccountInfo,
        last_bill_amount: float = 0,
    ) -> None:
        """
        Привязка лицевого счета к аккаунту личного кабинета.

        Параметры:
        - `account`: номер или публичная информация о лицевом счете
        - `last_bill_amount`: сумма последнего начисления.
        Может быть взята автоматически из публичной информации о счете.
        """

        if isinstance(account, PublicAccountInfo):
            last_bill_amount = last_bill_amount or account.balance
            account = account.account

        if account in self.accounts:
            _LOGGER.info("Лицевой счет %d уже привязан к аккаунту", account)
            return

        if last_bill_amount <= 0:
            raise ValueError("Сумма последнего начисления не указана")

        _LOGGER.debug("Привязка лицевого счета %d", account)

        async with self._post(
            "account/add", account=account, summ=last_bill_amount
        ) as x:
            html = await x.text()

        self._update_accounts(html)

        if account not in self.accounts:
            raise AccountBindingError("Не удалось привязать лицевой счет %d", account)

    @api(auth_required=True)
    async def account_rm(self, account: int) -> None:
        """
        Отвязка лицевого счета от аккаунта личного кабинета.

        Параметры:
        - `account`: номер лицевого счета.
        """

        if account not in self.accounts:
            _LOGGER.info("Лицевой счет %d не привязан к аккаунту", account)
            return

        async with self._post(f"account/{account}/remove") as x:
            self._update_accounts(await x.text())

        if account in self.accounts:
            raise AccountBindingError("Не удалось отвязать лицевой счет %d", account)

    async def _set_meters_values(
        self,
        path: str,
        values: Mapping[int, float],
    ) -> None:
        if not values:
            return

        async with self._get(path) as x:
            html = await x.text()

        data: dict[str, Any] = {}
        meters = parse_meters(html)

        # Если используем без авторизации - извлечем номер лицевого счета
        # из пути запроса и добавим в данные запроса
        if not path.startswith("account"):
            data["ls"] = int(path.rsplit("/", 1)[-1])

        for id, value in values.items():
            if m := meters.get(id):
                if value > m.value:
                    data[f"counters[{id}_0][value]"] = value
                    data[f"counters[{id}_0][rawId]"] = id
                    data[f"counters[{id}_0][tarif]"] = 0

                    continue

                raise ValueError(
                    f"Новое значение счетчика {id} должно быть выше текущего {m.value}"
                )

            raise ValueError(f"Счетчик {id} не найден")

        async with self._post(path, **data):
            pass

    @api(auth_required=True)
    async def meters_info(
        self, account: int | None = None
    ) -> Mapping[int, PublicMeterInfo]:
        """
        Запрос информации о приборах учета по лицевому счету.

        Возвращает словарь `идентификатор - информация о приборе учета`.

        Включает следующую информацию:
        - Внутренний идентификатор (для отправки новых показаний)
        - Серийный номер
        - Дата последнего показания
        - Последнее показание
        """

        async with self._get(f"account/{self._account(account)}/counters") as x:
            return parse_meters(await x.text())

    @api(public=True)
    async def pub_meters_info(self, account: int) -> Mapping[int, PublicMeterInfo]:
        """
        Запрос публичной информации о приборах учета по лицевому счету.

        Возвращает словарь `идентификатор - информация о приборе учета`.

        Включает следующую информацию:
        - Внутренний идентификатор (для отправки новых показаний)
        - Серийный номер
        - Дата последнего показания
        - Последнее показание

        Параметры:
        - `account`: номер лицевого счета.
        """

        async with self._get(f"counters/{account}") as x:
            return parse_meters(await x.text())

    @api(public=True)
    async def pub_set_meters_values(
        self,
        account: int,
        values: Mapping[int, float],
    ) -> None:
        """
        Передача новых показаний приборов учета без авторизации.

        Параметры:
        - `account`: номер лицевого счета.
        - `values`: словарь `идентификатор прибора - новое показание`.
        """

        await self._set_meters_values(f"counters/{account}", values)

    @api(public=True)
    async def pub_account_info(self, account: int) -> PublicAccountInfo | None:
        """
        Запрос открытой информации по лицевому счету.

        Параметры:
        - `account`: номер лицевого счета.
        """

        async with self._get("payment/checkLS", ls=account) as x:
            json: Mapping[str, Any] = await x.json()

        if json["checkLS"]:
            return PublicAccountInfo(
                account,
                str_normalize(json["address"]),
                to_float(json["balanceSumma"]),
                to_float(json["balancePeni"]),
            )

        _LOGGER.info("Лицевой счет %d не найден", account)

    @api(public=True)
    async def pub_accounts_info(
        self, *accounts: int
    ) -> Mapping[int, PublicAccountInfo]:
        """
        Запрос открытой информации по лицевым счетам.

        Параметры:
        - `accounts`: номера лицевых счетов.
        """

        result = await asyncio.gather(*map(self.pub_account_info, accounts))

        return {x.account: x for x in result if x}



import dataclasses as dc
import itertools as it
import logging
import re
from types import MappingProxyType
from typing import Any, cast

from bs4 import BeautifulSoup, Tag

from .account import AccountInfo
from .meters import PublicMeterInfo
from .utils import str_normalize, str_to_date

_LOGGER = logging.getLogger(__name__)

_RE_ACCOUNT_URL = re.compile(r"/\d+$")
_RE_RAWID = re.compile(r"rowId")


def parse_accounts(html: str) -> tuple[int, ...]:
    bs, ids = BeautifulSoup(html, "html.parser"), []
    tag = cast(Tag, bs.find("div", {"id": "select_ls_dropdown"}))

    for x in tag.find_all("a", {"href": _RE_ACCOUNT_URL}):
        ids.append(cast(str, x["href"]).rsplit("/", 1)[1])

    _LOGGER.debug(f"Лицевые счета: {", ".join(ids)}")

    return tuple(map(int, ids))


def parse_token(html: str) -> str:
    bs = BeautifulSoup(html, "html.parser")
    tag = cast(Tag, bs.find("meta", {"name": "csrf-token"}))
    token = cast(str, tag["content"])

    _LOGGER.debug("CSRF токен: %s", token)

    return token


def parse_account(html: str) -> AccountInfo:
    bs = BeautifulSoup(html, "html.parser")
    wl = cast(Tag, bs.find("div", class_="widget-left"))

    ws1 = cast(Tag, wl.find("div", class_="widget-section1"))
    ws1 = cast(Tag, ws1.find_all("div", class_="text-col-left"))

    ws2 = cast(Tag, wl.find("div", class_="widget-section2"))
    ws2 = cast(list[Tag], ws2.find_all("div", class_="text-col-right"))

    ws = (str_normalize(x.text) for x in it.chain(ws1, ws2))

    def _cnv(k, v):
        return k.type(v) if v != "-" else 0

    data: Any = (_cnv(*x) for x in zip(dc.fields(AccountInfo), ws))

    return AccountInfo(*data)


def parse_meters(html: str) -> MappingProxyType[int, PublicMeterInfo]:
    """
    Парсит HTML страницу с информацией по приборам учета.

    Возвращает словарь `идентификатор - информация о приборе учета`.
    """

    result: dict[int, PublicMeterInfo] = {}

    bs = BeautifulSoup(html, "html.parser")
    form = cast(Tag, bs.find("form", id="sendCountersValues"))

    for meter in form.find_all("div", class_="block-sch"):
        meter = cast(Tag, meter)

        name = cast(Tag, meter.find("span", class_="type"))

        if not name.text:
            continue

        serial = cast(Tag, name.find_next("span"))
        date = cast(Tag, meter.find(class_="block-note"))
        value = cast(Tag, date.find_next_sibling())

        name, serial = name.text, serial.text.rsplit("№", 1)[-1]
        date = str_to_date(date.text.strip().removeprefix("от "))
        value = float(value.text.strip())

        id = cast(Tag, meter.find("input", {"name": _RE_RAWID}))
        id = int(cast(str, id["value"]))

        result[id] = PublicMeterInfo(name, serial, date, value)

    return MappingProxyType(result)
